import { Trie } from './utils/trie';
const EQTB_HASH_SIZE = 19;
const EQTB_HASH_PRIME = 2 ** EQTB_HASH_SIZE - 1;
const EQTB_BASE_HASH_PRIME = 2 ** 8 + 1;
const PARENT_HASH_SIZE = 5;
const PARENT_HASH_PRIME = 2 ** PARENT_HASH_SIZE - 1;
const HASH_SIZE = 2 ** 31 - 1;
const MAX_DEPTH = 6;
export default class Context {
    constructor(groupType, depth = 0, parent) {
        this.groupType = groupType;
        this.depth = depth;
        this.parent = parent;
        this.reset = false;
        this.command = null;
        this.#hash = 0;
        this.#changed = true;
        if (depth !== undefined && parent !== undefined) {
            this.eqtb = { catcode: [] };
            return;
        }
        this.eqtb = {
            catcode: new Uint8Array(2 ** 16 * 6).fill(12 + (12 << 4)),
        };
    }
    #hash;
    #changed;
    get hash() {
        if (this.#changed) {
            this.#changed = false;
            this.#hash = this.generateHash();
        }
        return this.#hash;
    }
    defineCatCode(chr, cmd) {
        if (chr >= 0x40000 && chr < 0xe0000) {
            throw new RangeError('Unicode point is valid, but unassigned');
        }
        if (chr > 0x10ffff) {
            throw new RangeError('Unicode point is invalid');
        }
        if (chr > 0xdffff) {
            chr -= 0xe0000 - 0x40000;
        }
        const i = Math.floor(chr / 2);
        if (chr % 2 === 0) {
            this.eqtb.catcode[i] &= 0b11110000;
            this.eqtb.catcode[i] += cmd;
        }
        else {
            this.eqtb.catcode[i] &= 0b00001111;
            this.eqtb.catcode[i] += cmd << 4;
        }
        this.#changed = true;
    }
    getCatCode(ch) {
        if (ch >= 0x40000 && ch < 0xe0000) {
            throw new RangeError('Unicode point is valid, but unassigned');
        }
        if (ch > 0x10ffff) {
            throw new RangeError('Unicode point is invalid');
        }
        if (ch < 0) {
            return -1;
        }
        if (ch > 0xdffff) {
            ch -= 0xe0000 - 0x40000;
        }
        const i = Math.floor(ch / 2);
        let n = this;
        while (n && n.eqtb.catcode[i] === undefined) {
            n = n.parent;
        }
        if (n === undefined) {
            throw new Error("this shouldn't happen");
        }
        if (ch % 2 === 0) {
            return n.eqtb.catcode[i] & 0b00001111;
        }
        return n.eqtb.catcode[i] >> 4;
    }
    generateEqTbHash() {
        let b = EQTB_BASE_HASH_PRIME;
        return (this.eqtb.catcode.reduce((h, v) => {
            b = (b * EQTB_BASE_HASH_PRIME) % EQTB_HASH_PRIME;
            return (h + v * b) % EQTB_HASH_PRIME;
        }, 0) <<
            (PARENT_HASH_SIZE + MAX_DEPTH));
    }
    generateHash() {
        let h = (this.parent ? this.parent.hash : 0) % PARENT_HASH_PRIME;
        h += this.depth << PARENT_HASH_SIZE;
        h += this.generateEqTbHash();
        h %= HASH_SIZE;
        return h;
    }
}
Context.primitives = new Trie(null);

//# sourceMappingURL=context.js.map
