import { ExternalTokenizer } from 'lezer';
import { catcode } from './enums/catcode';
import Context from './context';
import { control_sequence_token, Dialect_directives, directive_comment, line_comment, } from './gen/terms';
import cp from './utils/c';
import isHex from './utils/is-hex';
class State {
    constructor(spec) {
        this.chr = spec.chr;
        this.loc = spec.loc;
        this.buf = spec.buf;
        this.tok = spec.tok;
        this.ctx = spec.ctx;
        this.dct = spec.dct;
    }
}
export default class Tokenizer extends ExternalTokenizer {
    constructor() {
        super((buf, tok, stk) => {
            if (tok.start === 0) {
                stk.context.reset = true;
            }
            if (tok.start >= buf.length) {
                return undefined;
            }
            let dct = 0;
            if (stk.dialectEnabled(Dialect_directives)) {
                dct |= 1;
            }
            this.#state = new State({
                chr: buf.get(tok.start),
                loc: tok.start + 1,
                buf,
                tok,
                ctx: stk.context,
                dct,
            });
            return this.getNext();
        }, { contextual: true });
        this.#offset = 2;
        this.nextIsDirective = () => {
            return this.#state.buf.get(this.#state.loc) === cp `!`;
        };
    }
    #state;
    #offset;
    getNext() {
        this.#state.cmd = this.#state.ctx.getCatCode(this.#state.chr);
        switch (this.#state.cmd) {
            case 1:
            case 2:
            case 3:
            case 4:
            case 5:
            case 6:
            case 8:
            case 9:
            case 10:
            case 11:
            case 12:
            case 13:
            case 15: {
                this.#state.tok.accept(catcode[this.#state.cmd], this.#state.loc);
                break;
            }
            case 0: {
                this.scanControlSequence();
                const cs = Context.primitives.lookup(this.#state.cs);
                if (cs !== null) {
                    this.#state.tok.accept(cs[0], this.#state.loc);
                    break;
                }
                this.#state.tok.accept(control_sequence_token, this.#state.loc);
                break;
            }
            case 14: {
                if ((this.#state.dct & 1) > 0 && this.nextIsDirective()) {
                    this.scanComment();
                    this.#state.tok.accept(directive_comment, this.#state.loc);
                    break;
                }
                this.scanComment();
                this.#state.tok.accept(line_comment, this.#state.loc);
                break;
            }
            case 7: {
                if (this.nextIsExpandedCharacter()) {
                    this.scanExpandedCharacter();
                    this.#state.loc += this.#offset;
                    this.getNext();
                    break;
                }
                this.#state.tok.accept(catcode[this.#state.cmd], this.#state.loc);
                break;
            }
            default: {
                throw new Error('Unknown character');
            }
        }
    }
    scanControlSequence() {
        if (this.#state.loc >= this.#state.buf.length) {
            return;
        }
        this.#state.chr = this.#state.buf.get(this.#state.loc++);
        this.#state.cmd = this.#state.ctx.getCatCode(this.#state.chr);
        const cs = [this.#state.chr];
        if (this.#state.cmd === 7 && this.nextIsExpandedCharacter()) {
            this.scanExpandedCharacter();
            this.#state.loc += this.#offset;
            this.#state.cmd = this.#state.ctx.getCatCode(this.#state.chr);
            cs[0] = this.#state.chr;
        }
        if (this.#state.cmd !== 11) {
            this.#state.cs = String.fromCodePoint(...cs);
            return;
        }
        do {
            this.#state.chr = this.#state.buf.get(this.#state.loc++);
            this.#state.cmd = this.#state.ctx.getCatCode(this.#state.chr);
            cs.push(this.#state.chr);
            if (this.#state.cmd === 7 && this.nextIsExpandedCharacter()) {
                this.scanExpandedCharacter();
                this.#state.cmd = this.#state.ctx.getCatCode(this.#state.chr);
                if (this.#state.cmd === 11) {
                    cs.push(this.#state.chr);
                    this.#state.loc += this.#offset;
                }
            }
        } while (this.#state.cmd === 11);
        this.#state.loc -= 1;
        cs.pop();
        this.#state.cs = String.fromCodePoint(...cs);
    }
    scanComment() {
        do {
            this.#state.chr = this.#state.buf.get(this.#state.loc++);
        } while (this.#state.chr > -1 &&
            this.#state.ctx.getCatCode(this.#state.chr) !== 5);
    }
    nextIsExpandedCharacter() {
        if (this.#state.ctx.getCatCode(this.#state.buf.get(this.#state.loc)) !== 7) {
            return false;
        }
        const c = this.#state.buf.get(this.#state.loc + 1);
        return c > 0 && c < 0o200;
    }
    scanExpandedCharacter() {
        this.#offset = 2;
        this.#state.chr = this.#state.buf.get(this.#state.loc + 1);
        if (isHex(this.#state.chr)) {
            const cc = this.#state.buf.get(this.#state.loc + 2);
            if (isHex(cc)) {
                this.#offset += 1;
                this.#state.chr = parseInt(`0x${String.fromCharCode(this.#state.chr, cc)}`, 16);
            }
        }
        this.#state.chr = this.#state.chr < 0o100 ? this.#state.chr + 0o100 : this.#state.chr - 0o100;
    }
}

//# sourceMappingURL=tokenizer.js.map
